#include <functional>
#include <iostream>
#include <pthread.h>
#include <sys/syscall.h>
#include <getopt.h>
#include <chrono>
#include <asm/types.h>
#include <assert.h>
#include <fcntl.h>
#include <dirent.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <linux/stddef.h>
#include <linux/videodev2.h>
#include <errno.h>

#include <common.hpp>

#define FMT_NUM_PLANES	1
#define pixfmtstr(x) (x) & 0xff, ((x) >> 8) & 0xff, ((x) >> 16) & 0xff, ((x) >> 24) & 0xff

static void v4l2_camera_free_buffers(DrmBuffer **buffers, int n_buffer, int drm_fd, bool m_cmabuffer)
{
	int i;

	for (i = 0; i < n_buffer; i++) {
		if (buffers[i]) {
			if (m_cmabuffer)
				RockchipCmaFree(drm_fd, buffers[i]);
			else
				RockchipCmaFree(drm_fd, buffers[i]);
		}
		buffers[i] = NULL;
	}
}

static int v4l2_camera_alloc_buffers(DrmBuffer **buffers, int n_buffer, int drm_fd,
		__u32 format, __u32 width, __u32 height, bool m_cmabuffer)
{
	int i;

	for (i = 0; i < n_buffer; i++) {
		if (m_cmabuffer)
			buffers[i] = RockchipCmaAlloc(drm_fd, format, width, height);
		else
			buffers[i] = RockchipDrmAlloc(drm_fd, format, width, height);

		if (buffers[i] == NULL)
			goto exit;
	}
	
	return 0;

exit:
	v4l2_camera_free_buffers(buffers, n_buffer, drm_fd, m_cmabuffer);
	return -ENOMEM;
}

static int v4l2_camera_open(int dev)
{
	int fd;
	char v4l2_dev[32];
	sprintf(v4l2_dev, "/dev/video%d", dev);

	fd = open(v4l2_dev, O_RDWR, 0);
	if(fd < 0)
		fprintf(stderr, "Open dev %s failed\n", v4l2_dev);
	return fd;
}

static int v4l2_camera_setfmt(int fd, enum v4l2_buf_type type, __u32 width, __u32 height, __u32 format)
{
	int ret, found = 0;
	struct v4l2_format fmt;
	struct v4l2_fmtdesc fmtdesc;

	fmtdesc.index = 0;
	fmtdesc.type = type;
	while(1) {
		ret = ioctl(fd, VIDIOC_ENUM_FMT, &fmtdesc);
		if(ret < 0)
			break;

		if(fmtdesc.pixelformat == format) {
			found = 1;
			break;
		}
		fmtdesc.index++;
	}

	if(found == 0) {
		fprintf(stderr, "This format(%c%c%c%c) is NOT supported\n", pixfmtstr(format));
		return -EINVAL;
	}

	memset(&fmt, 0, sizeof(fmt));
	fmt.type = type;
	fmt.fmt.pix.width = width;
	fmt.fmt.pix.height = height;
	fmt.fmt.pix.pixelformat = format;
	fmt.fmt.pix.field = V4L2_FIELD_ANY;//V4L2_FIELD_INTERLACED;
	ret = ioctl(fd, VIDIOC_S_FMT, &fmt);
	if(ret < 0)
		fprintf(stderr, "Set format failed\n");

	return ret;
}

static int v4l2_camera_querycap(int fd)
{
	int ret;
	struct v4l2_capability cap;

	ret = ioctl(fd, VIDIOC_QUERYCAP, &cap);
	if(ret < 0) {
		fprintf(stderr, "Querycap failed\n");
		return ret;
	}

	if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) && !(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE)) {
		fprintf(stderr, "The device is NOT video capture device\n");
		return -ENOSTR;
	}

	if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
		fprintf(stderr, "The device is NOT video streaming io\n");
		return -ENOSTR;
	}

	return 0;
}

static int v4l2_camera_reqbufs(int fd, enum v4l2_buf_type type, __u32 memory, int n_buffer)
{
	int ret;
	struct v4l2_requestbuffers req;

	memset(&req, 0, sizeof(req));
	req.count = n_buffer;
	req.type = type;
	req.memory = memory;
	ret = ioctl(fd, VIDIOC_REQBUFS, &req);
	if(ret < 0) 
		fprintf(stderr, "Request buffers failed\n");
	
	return ret;
}

static int v4l2_camera_enqueue(int fd, enum v4l2_buf_type type, __u32 memory, bool hasplane, DrmBuffer **buffers, int index)
{
	int ret, i;
	struct v4l2_buffer buf;
	struct v4l2_plane planes[FMT_NUM_PLANES];

	memset(&buf, 0, sizeof(buf));
	for (i = 0; i < FMT_NUM_PLANES; i++)
		memset(&planes[i], 0, sizeof(struct v4l2_plane));

	buf.type = type;
	buf.memory = memory;
	buf.index = index;
	
	if (hasplane) {
		switch (memory) {
		case V4L2_MEMORY_USERPTR:
			planes[0].m.userptr = (unsigned long)buffers[index]->virt;
			break;
		case V4L2_MEMORY_DMABUF:
			planes[0].m.fd = buffers[index]->fd;
			break;
		default:
			fprintf(stderr, "Unsupported memroy type: %u", memory);
			return -EINVAL;
		}
		planes[0].length = buffers[index]->size;
		buf.m.planes = planes;
		buf.length = FMT_NUM_PLANES;
	} else {
		switch (memory) {
		case V4L2_MEMORY_USERPTR:
			buf.m.userptr = (unsigned long)buffers[index]->virt;
			break;
		case V4L2_MEMORY_DMABUF:
			planes[0].m.fd = buffers[index]->fd;
			break;
		default:
			fprintf(stderr, "Unsupported memroy type: %u", memory);
			return -EINVAL;
		}
		buf.length = buffers[index]->size;
	}

	ret = ioctl(fd, VIDIOC_QBUF, &buf);
	if(ret < 0)
		fprintf(stderr, "Enqueue buffers failed, index %d\n", index);

	return ret;
}

static int v4l2_camera_dequeue(int fd, enum v4l2_buf_type type, __u32 memory, bool hasplane)
{
	int ret;
	struct v4l2_buffer buf;
	struct v4l2_plane planes[FMT_NUM_PLANES];

	memset(&buf, 0, sizeof(buf));
	buf.type = type;
	buf.memory = memory;

	if (hasplane) {
		buf.m.planes = planes;
		buf.length = FMT_NUM_PLANES;
	}

	ret = ioctl(fd, VIDIOC_DQBUF, &buf);
	if(ret < 0) {
		fprintf(stderr, "Dequeue buffer failed, ret = %d\n", ret);
		return ret;
	}

	return buf.index;
}

static int v4l2_camera_streamon(int fd, enum v4l2_buf_type type)
{
	int ret;
	enum v4l2_buf_type buf_type = type;
	ret = ioctl(fd, VIDIOC_STREAMON, &buf_type);
	if(ret < 0)
		fprintf(stderr, "Stream on failed\n");

	return ret;
}

static int v4l2_camera_streamoff(int fd, enum v4l2_buf_type type)
{
	int ret;
	enum v4l2_buf_type buf_type = type;
	ret = ioctl(fd, VIDIOC_STREAMOFF, &buf_type);
	if(ret < 0)
		fprintf(stderr, "Stream off failed\n");

	return ret;
}

v4l2_camera::v4l2_camera(int dev, CAM_TYPE type, __u32 width, __u32 height, bool is_cmabuffer)
{
	int i;

	switch (type) {
	case CAM_TYPE_USB:
		m_memory = V4L2_MEMORY_USERPTR;
		m_type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		m_format = V4L2_PIX_FMT_YUYV;
		m_hasplane = false;
		break;
	case CAM_TYPE_MIPI:
		m_memory = V4L2_MEMORY_DMABUF;
		m_type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
		m_format = V4L2_PIX_FMT_NV12;
		m_hasplane = true;
		break;
	case CAM_TYPE_HDMIRX:
		m_memory = V4L2_MEMORY_DMABUF;
		m_type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
		m_format = V4L2_PIX_FMT_BGR24;
		m_hasplane = true;
		break;
	default:
		break;
	}

	m_dev = dev;
	m_width = width;
	m_height = height;
	drm_fd = -1;
	v4l2_fd = -1;
	
	n_buffer = sizeof(buffers) / sizeof(buffers[0]);
	for (i = 0; i < n_buffer; i++)
		buffers[i] = NULL;	
	m_cmabuffer = is_cmabuffer;
	m_current = -1;

	m_streamon = false;
}

v4l2_camera::~v4l2_camera(void)
{
	if (m_streamon) {
		assert(v4l2_camera_streamoff(v4l2_fd, m_type) >= 0);
		m_streamon = false;
	}

	v4l2_camera_free_buffers(buffers, n_buffer, drm_fd, m_cmabuffer);

	if (drm_fd >= 0)
		RockchipDrmClose(drm_fd);

	if (v4l2_fd >= 0)
		close(v4l2_fd);
}

int v4l2_camera::start(void)
{
	int ret = 0, i;

	drm_fd = RockchipDrmOpen();
	if (drm_fd < 0) {
		ret = drm_fd;
		goto exit;
	}

	ret = v4l2_camera_alloc_buffers(buffers, n_buffer, drm_fd, m_format, m_width, m_height, m_cmabuffer);
	if (ret < 0)
		goto exit;

	v4l2_fd = v4l2_camera_open(m_dev);
	if (v4l2_fd < 0) {
		ret = v4l2_fd;
		goto exit;
	}

	ret = v4l2_camera_querycap(v4l2_fd);
	if (ret < 0)
		goto exit;

	ret = v4l2_camera_setfmt(v4l2_fd, m_type, m_width, m_height, m_format);
	if (ret < 0)
		goto exit;

	ret = v4l2_camera_reqbufs(v4l2_fd, m_type, m_memory, n_buffer); 
	if (ret < 0)
		goto exit;

	for (i = 0; i < n_buffer; i++) {
		ret = v4l2_camera_enqueue(v4l2_fd, m_type, m_memory, m_hasplane, buffers, i);
		if (ret < 0)
			goto exit;
	}

	ret = v4l2_camera_streamon(v4l2_fd, m_type);
	if (ret < 0)
		goto exit;

	m_streamon = true;
	return 0;
exit:
	if (m_streamon) {
		assert(v4l2_camera_streamoff(v4l2_fd, m_type) >= 0);
		m_streamon = false;
	}

	v4l2_camera_free_buffers(buffers, n_buffer, drm_fd, m_cmabuffer);

	if (v4l2_fd >= 0) {
		close(v4l2_fd);
		v4l2_fd = -1;
	}

	if (drm_fd >= 0) {
		RockchipDrmClose(drm_fd);
		drm_fd = -1;
	}

	return ret;
}

DrmBuffer *v4l2_camera::get_frame(void)
{
	int ret;

	if (m_current >= 0) {
		ret = v4l2_camera_enqueue(v4l2_fd, m_type, m_memory, m_hasplane, buffers, m_current);
		if (ret < 0)
			return NULL;
		m_current = -1;
	}

	ret = v4l2_camera_dequeue(v4l2_fd, m_type, m_memory, m_hasplane);
	if (ret < 0)
		return NULL;

	m_current = ret;
	return buffers[ret];
}
